// Copyright 2019-2020 Francesco Biscani (bluescarni@gmail.com)
//
// This file is part of the obake library.
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

#ifndef OBAKE_CONFIG_HPP
#define OBAKE_CONFIG_HPP

// Start of defines instantiated by CMake.
// clang-format off
#define OBAKE_VERSION_STRING "0.8.0"
#define OBAKE_VERSION_MAJOR 0
#define OBAKE_VERSION_MINOR 8
#define OBAKE_VERSION_PATCH 0

// clang-format on
// End of defines instantiated by CMake.

// Create a shortcut for detecting GCC.
// NOTE: clang defines __GNUC__ internally.
#if defined(__GNUC__) && !defined(__clang__)

#define OBAKE_COMPILER_IS_GCC

#endif

// Detect 128bit integers, lifted from mp++. This is part of the public API.
#include <mp++/config.hpp>

#if defined(MPPP_HAVE_GCC_INT128)

#define OBAKE_HAVE_GCC_INT128

#endif

// likely/unlikely macros, supported on GCC/clang.
#if defined(__GNUC__) || defined(__clang__)

#define obake_likely(x) __builtin_expect(static_cast<bool>(x), 1)
#define obake_unlikely(x) __builtin_expect(static_cast<bool>(x), 0)

#else

#define obake_likely(x) (x)
#define obake_unlikely(x) (x)

#endif

// For packing 64-bit values, we need 128-bit multiplication.
// This requires either 128-bit integer types, or compiler intrinsics.
// NOTE: _WIN64 should catch also 64-bit arm.
#if defined(OBAKE_HAVE_GCC_INT128) || (defined(_MSC_VER) && defined(_WIN64))

#define OBAKE_PACKABLE_INT64

#endif

#endif
